/******************************************************************
 Project Name: Terminal Tetris
 File Name: network.c
 Description: Ʈũ   Լ ִ .
 
 Created: 2004. 12. 22
 RCS-ID: $Id: network.c,v 1.37 2004/06/11 05:30:13 myunggoni Exp $
 Copyright: 2004 (c) by myunggoni
 License: GNU General Public License
 Author: Myung-gon, Park <myunggoni@users.kldp.net>
 *******************************************************************/
#include <ncurses.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include "main.h"
#include "config.h"

extern const char blocks[7][4][4][4];
extern char board[MAX_HEIGHT][MAX_WIDTH];
extern int start_game;
extern int score;
extern int level;
int network_game;
int sock;
int game_result;
pthread_mutex_t mutex_lock = PTHREAD_MUTEX_INITIALIZER;

int server_mode() { 
	int serv_sock;
	int option;
	int ch;
	char *rval;
	unsigned short int port;
	struct sockaddr_in serv_addr;
	struct sockaddr_in clnt_addr;
	socklen_t clnt_addr_size;
	pthread_t main_thread;
	pthread_t rcv_thread;
	void *main_result;
	void *rcv_result;
	
	curs_set(1);
	echo();
		
	rval = get_value("Enter Your Port Number(Default 8161): "); 	
	
	if(is_number(rval)) 
		port = atoi(rval);
	else 
		port = DEFAULT_PORT;
	
	do {	
		if((serv_sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) 
			error_handling("socket");

		option = 1;

		if(setsockopt(serv_sock, SOL_SOCKET, SO_REUSEADDR, &option, sizeof(option)) < 0)
			error_handling("setsocket");

		memset(&serv_addr, 0, sizeof(serv_addr));
		serv_addr.sin_family = AF_INET;
		serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		serv_addr.sin_port = htons(port);

		if(bind(serv_sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
			error_handling("bind");
		}

		clear();
		curs_set(0);
		mvprintw(MENU_Y, MENU_X, "Waiting Client...");
		refresh();

		if(listen(serv_sock, 5) < 0) 
			error_handling("listen");


		clnt_addr_size = sizeof(clnt_addr);
		if((sock = accept(serv_sock, (struct sockaddr *)&clnt_addr, &clnt_addr_size)) < 0) 
			error_handling("accept");

		if(close(serv_sock))
			error_handling("close");
		
		mvprintw(MENU_Y + 2, MENU_X, "Connected from %s", inet_ntoa(clnt_addr.sin_addr));
		network_game = 1;
		level = 1;
		refresh();
		sleep(2);

		if(pthread_create(&main_thread, NULL, network_main, NULL)) 
			error_handling("pthread_create");

		if(pthread_create(&rcv_thread, NULL, network_receiver, NULL)) 
			error_handling("pthread_create");

		if(pthread_join(main_thread, &main_result))
			error_handling("pthread_join");

		if(pthread_join(rcv_thread, &rcv_result))
			error_handling("pthread_join");
		
		if(close(sock))
			error_handling("close");

		start_game = 0;
		
		if(game_result == 1)
			game_over(NETWORK_WIN); 
		else if(game_result == -1)
			game_over(NETWORK_LOSE); 
		else
			game_over(NETWORK_DRAW); 
		
		clear();
		mvprintw(MENU_Y, MENU_X, "Play Again?(y/n): ");
		ch = getch();

		while(ch != 'y' && ch != 'Y' && ch != 'n' && ch != 'N') {
			mvhline(MENU_Y, MENU_X, ' ', 20);
			mvprintw(MENU_Y, MENU_X, "Wrong Input. \'y\' or \'n\': ");

			ch = getch();
		}

	} while(ch == 'y' || ch == 'Y');

	return 0; 
}

int client_mode() { 
	char *rval;
	unsigned short int port;
	int ch;
	struct sockaddr_in serv_addr;
	pthread_t main_thread;
	pthread_t rcv_thread;
	void *main_result;
	void *rcv_result;

	curs_set(1);
	echo();

	rval = get_value("Enter Your Port Number(Default 8161): "); 

	
	if(is_number(rval)) 
		port = atoi(rval);
	else 
		port = DEFAULT_PORT;
	
	rval = get_value("Enter Your Tetris Server IP: "); 

	do {
		if((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) 
			error_handling("socket");

		memset(&serv_addr, 0, sizeof(serv_addr));
		serv_addr.sin_family = AF_INET;
		serv_addr.sin_port = htons(port);

		if(inet_pton(AF_INET, rval, &serv_addr.sin_addr) < 0)
			error_handling("inet_pton");

		clear();
		curs_set(0);
		mvprintw(MENU_Y, MENU_X, "Connecting to %s:%d", rval, port);
		
		if(connect(sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) 
			error_handling("connect");

		mvprintw(MENU_Y + 2, MENU_X, "Connected!");
		network_game = 1;
		level = 1;
		refresh();
		sleep(2);

		if(pthread_create(&main_thread, NULL, network_main, NULL)) 
			error_handling("pthread_create");

		if(pthread_create(&rcv_thread, NULL, network_receiver, NULL)) 
			error_handling("pthread_create");

		if(pthread_join(main_thread, &main_result))
			error_handling("pthread_join");

		if(pthread_join(rcv_thread, &rcv_result))
			error_handling("pthread_join");
		
		if(close(sock) < 0)
			error_handling("close");

		start_game = 0;
		
		if(game_result == 1)
			game_over(NETWORK_WIN); 
		else if(game_result == -1)
			game_over(NETWORK_LOSE); 
		else
			game_over(NETWORK_DRAW); 
		
		clear();
		mvprintw(MENU_Y, MENU_X, "Play Again? Server Must to Ready in Advance(y/n): ");
		ch = getch();

		while(ch != 'y' && ch != 'Y' && ch != 'n' && ch != 'N') {
			mvhline(MENU_Y, MENU_X, ' ', 55);
			mvprintw(MENU_Y, MENU_X, "Wrong Input. \'y\' or \'n\': ");

			ch = getch();
		}
		
	} while(ch == 'y' || ch == 'Y');

	return 0; 
}

int server_mode_direct(unsigned short int port) { 
	int serv_sock;
	int option;
	int ch;
	struct sigaction sa;
	struct sockaddr_in serv_addr;
	struct sockaddr_in clnt_addr;
	socklen_t clnt_addr_size;
	pthread_t main_thread;
	pthread_t rcv_thread;
	void *main_result;
	void *rcv_result;
	
	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = signal_handler;
	sigaction(SIGALRM, &sa, NULL);

	initscr();
	curs_set(0);
	cbreak();	
	keypad(stdscr, TRUE);	
	srand(time(NULL));

	do {	
		if((serv_sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) 
			error_handling("socket");

		option = 1;

		if(setsockopt(serv_sock, SOL_SOCKET, SO_REUSEADDR, &option, sizeof(option)) < 0)
			error_handling("setsocket");

		memset(&serv_addr, 0, sizeof(serv_addr));
		serv_addr.sin_family = AF_INET;
		serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		serv_addr.sin_port = htons(port);

		if(bind(serv_sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
			error_handling("bind");
		}

		clear();
		mvprintw(MENU_Y, MENU_X, "Waiting Client...");
		refresh();

		if(listen(serv_sock, 5) < 0) 
			error_handling("listen");


		clnt_addr_size = sizeof(clnt_addr);
		if((sock = accept(serv_sock, (struct sockaddr *)&clnt_addr, &clnt_addr_size)) < 0) 
			error_handling("accept");

		if(close(serv_sock))
			error_handling("close");
		
		mvprintw(MENU_Y + 2, MENU_X, "Connected from %s", inet_ntoa(clnt_addr.sin_addr));
		network_game = 1;
		level = 1;
		refresh();
		sleep(2);

		if(pthread_create(&main_thread, NULL, network_main, NULL)) 
			error_handling("pthread_create");

		if(pthread_create(&rcv_thread, NULL, network_receiver, NULL)) 
			error_handling("pthread_create");

		if(pthread_join(main_thread, &main_result))
			error_handling("pthread_join");

		if(pthread_join(rcv_thread, &rcv_result))
			error_handling("pthread_join");
		
		if(close(sock))
			error_handling("close");
	
		start_game = 0;
		
		if(game_result == 1)
			game_over(NETWORK_WIN); 
		else if(game_result == -1)
			game_over(NETWORK_LOSE); 
		else
			game_over(NETWORK_DRAW); 
		
		clear();
		mvprintw(MENU_Y, MENU_X, "Play Again?(y/n): ");
		ch = getch();

		while(ch != 'y' && ch != 'Y' && ch != 'n' && ch != 'N') {
			mvhline(MENU_Y, MENU_X, ' ', 20);
			mvprintw(MENU_Y, MENU_X, "Wrong Input. \'y\' or \'n\': ");

			ch = getch();
		}

	} while(ch == 'y' || ch == 'Y');

	return 0; 
}

int client_mode_direct(char *server_address, unsigned short int port) { 
	int ch;
	struct sigaction sa;
	struct sockaddr_in serv_addr;
	pthread_t main_thread;
	pthread_t rcv_thread;
	void *main_result;
	void *rcv_result;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = signal_handler;
	sigaction(SIGALRM, &sa, NULL);

	initscr();
	curs_set(0);
	cbreak();	
	keypad(stdscr, TRUE);	
	srand(time(NULL));

	do {
		if((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) 
			error_handling("socket");

		memset(&serv_addr, 0, sizeof(serv_addr));
		serv_addr.sin_family = AF_INET;
		serv_addr.sin_port = htons(port);

		if(inet_pton(AF_INET, server_address, &serv_addr.sin_addr) < 0)
			error_handling("inet_pton");

		clear();
		mvprintw(MENU_Y, MENU_X, "Connecting to %s:%d", server_address, port);
		
		if(connect(sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) 
			error_handling("connect");

		mvprintw(MENU_Y + 2, MENU_X, "Connected!");
		network_game = 1;
		level = 1;
		refresh();
		sleep(2);
		
		if(pthread_create(&main_thread, NULL, network_main, NULL)) /*  κ. */
			error_handling("pthread_create");

		if(pthread_create(&rcv_thread, NULL, network_receiver, NULL)) /* ޱ⸸ Ѵ. */
			error_handling("pthread_create");

		if(pthread_join(main_thread, &main_result))
			error_handling("pthread_join");

		if(pthread_join(rcv_thread, &rcv_result))
			error_handling("pthread_join");

		if(close(sock) < 0)
			error_handling("close");

		start_game = 0;

		if(game_result == 1)
			game_over(NETWORK_WIN); 
		else if(game_result == -1)
			game_over(NETWORK_LOSE); 
		else
			game_over(NETWORK_DRAW); 
	
		clear();
		mvprintw(MENU_Y, MENU_X, "Play Again? Server Must to Ready in Advance(y/n): ");
		ch = getch();

		while(ch != 'y' && ch != 'Y' && ch != 'n' && ch != 'N') {
			mvhline(MENU_Y, MENU_X, ' ', 55);
			mvprintw(MENU_Y, MENU_X, "Wrong Input. \'y\' or \'n\': ");

			ch = getch();
		}
		
	} while(ch == 'y' || ch == 'Y');

	return 0; 
}

void *network_main(void *arg) { 
	int ch;
	struct itimerval val;
	const size_t packet_len = sizeof(packet);
	packet data;
	
	start_game = 1;
	score = 0;
	game_result = 0;
	
	clear(); 
	noecho();
	curs_set(0);
	clear();
	init_board();
	draw_table();
	draw_board();
	draw_enemy_table();
	create_random_blocks();
	preview_blocks();
	draw_info();

	memset(&val, 0, sizeof(val));
	val.it_interval.tv_usec = (1.0 / level) * 1000000 ;
	val.it_value.tv_usec = (1.0 / level) * 1000000 ;
	setitimer(ITIMER_REAL, &val, NULL);
	
	while(1) {
		ch = getch();
		
		switch(ch) {
			case KEY_LEFT:
			move_blocks_left();		
			break;
			
			case KEY_RIGHT:
			move_blocks_right();			
			break;
			
			case KEY_DOWN:
			move_blocks_down();
			break;
			
			case KEY_UP:
			rotate_blocks();
			break;

			case ' ':
			move_blocks_down_direct();
			break;

		}

		if(start_game == 0) { /*  . */
			memset(&data, 0, packet_len);
			data.status = htonl(NETWORK_LOSE);
		
			if(pthread_mutex_lock(&mutex_lock) < 0)
				error_handling("pthread_mutex_lock");
			
			if(send(sock, &data, packet_len, 0) < 0)
				error_handling("send");
			
			if(pthread_mutex_unlock(&mutex_lock) < 0)
				error_handling("pthread_mutex_unlock");
			
			break;
		} else if(network_game == 0) { /*  . */
			break;
		} else {
			preview_blocks();
			draw_info();
		}
	}


	
	return NULL; 
}

void *network_receiver(void *arg) {
	int i, j, k, count;
	unsigned int received;
	unsigned int total_received;
	unsigned int attack_lines;
	int status;
	const size_t packet_len = sizeof(packet);
	packet data;

	while(network_game) {
		total_received = 0;
	
		while(total_received < packet_len) {
			if((received = recv(sock, &data, packet_len, 0)) < 0)
				error_handling("recv");
		
			total_received += received;
		}
		
		status = ntohl(data.status);
		
		if(status) { /*  . */
			
			switch(status) {
				case NETWORK_LOSE : /*  ̰ . */
					memset(&data, 0, packet_len);
					data.status = htonl(NETWORK_WIN);

					if(pthread_mutex_lock(&mutex_lock) < 0)
						error_handling("pthread_mutex_lock");

					if(send(sock, &data, packet_len, 0) < 0)
						error_handling("send");

					if(pthread_mutex_unlock(&mutex_lock) < 0)
						error_handling("pthread_mutex_unlock");

					if(game_result == -1)
						game_result = 0;
					else
						game_result++;
					
					break;

				case NETWORK_WIN: /*   . */
					memset(&data, 0, packet_len);
					data.status = htonl(NETWORK_CONFIRM);

					if(pthread_mutex_lock(&mutex_lock) < 0)
						error_handling("pthread_mutex_lock");

					if(send(sock, &data, packet_len, 0) < 0)
						error_handling("send");

					if(pthread_mutex_unlock(&mutex_lock) < 0)
						error_handling("pthread_mutex_unlock");

					if(game_result == 1)
						game_result = 0;
					else
						game_result--;

					network_game = 0;
					break;

				case NETWORK_CONFIRM: /*  . */
					network_game = 0;
					break;
			}
			
		} else { /*   . */
			if((attack_lines = ntohl(data.attack_lines)) > 0) {

				count = 0;
				while(count < attack_lines) {
					check_attack();
					for(i = 1; i < MAX_HEIGHT - 1; i++) {
						for(j = 1; j < MAX_WIDTH - 1; j++) {
							board[i][j] = board[i + 1][j];
						}
					}
					
					k = (rand() % (MAX_WIDTH - 1)) + 1;
					board[MAX_HEIGHT - 2][k] = 0;
					draw_board();
					count++;
				}

			}
			
			draw_enemy_board(data);
		}
	}

	return NULL; 
}

